"""
service.py

Prediction service. 
"""
import uuid
import json
from importlib import import_module

from akerbp.mlops import __version__ as version
from akerbp.mlops.core import config, logger
import akerbp.mlops.cdf.helpers as mlops_helpers
from cognite.client.exceptions import CogniteNotFoundError

c = config.read_service_settings()
model_module = import_module(c.model_import_path)
predict = model_module.predict
_initialization = model_module.initialization
ModelException = model_module.ModelException

logging=logger.get_logger("mlops_services")

logging.debug(f"MLOps framework version {version}")


def initialization(secrets):
    """
    Read initialization object required by `predict`
    """
    # This check adds a little overhead to each prediction
    if 'init_object' not in globals():
        global init_object
        artifact_folder = c.artifact_folder
        init_object = _initialization(artifact_folder, secrets)
        logging.debug(f"Loaded initialization object")


def service(data, secrets):
    """
    Generate prediction for an input
    Inputs:
        data: input to the model (sent by a user through the API)
        secrets: api keys that the model may need during initialization
    Output:
        Dictionary with a status field ('ok' or 'error'). If status is 'ok', 
         there is a 'prediction' field as well. If status is 'error', there is 
         a 'message' field as well.
    """
    try:
        initialization(secrets)
        y = predict(data, init_object, secrets)
        if data["to_file"]: # make this more robust, include in data_validation. E.g. set to False if not specified in data dictionary 
            mlops_helpers.api_keys = secrets 
            mlops_helpers.set_up_cdf_client()
            cdf_client = mlops_helpers.global_client
            external_file_id = f"{c.model_name}_predictions_{uuid.uuid4().hex}.binary"
            try: 
                cdf_client["files"].files.delete(external_id=external_file_id)
            except CogniteNotFoundError:
                pass
            cdf_client["files"].files.upload_bytes(
                content=str(json.dumps(y)), name=f"{c.model_name}_predictions", external_id=external_file_id
            )
            logging.info(f"Prediction file uploaded to {external_file_id}")
            return dict(status="ok", prediction=False, prediction_file=external_file_id, model_id=c.model_id)
        else:
            return dict(status="ok", prediction=y, prediction_file=False, model_id=c.model_id)
    except ModelException as error:
        error_message = f"Could not get a prediction. Message: {error}"
        logging.error(error_message)
        return dict(status="error", message=error_message, model_id=c.model_id)



# handler.py
import traceback
from importlib import import_module
import warnings
warnings.simplefilter("ignore")

import akerbp.mlops.cdf.helpers as cdf
from akerbp.mlops.core import logger, config
service_name = config.envs.service_name
logging=logger.get_logger("mlops_cdf")

service = import_module(f"akerbp.mlops.services.{service_name}").service


def handle(data, secrets, function_call_info):
   try:
      if data:
         output = service(data, secrets)
      else:
         output = dict(status='ok')
      cdf.api_keys = secrets
      cdf.set_up_cdf_client(context='deploy')
      metadata = cdf.get_function_metadata(function_call_info["function_id"])
      output.update(dict(metadata=metadata))
      return output
   except Exception:
      trace = traceback.format_exc()
      error_message = f"{service_name} service failed.\n{trace}"
      logging.critical(error_message)
      return dict(status='error', error_message=error_message)



# helpers.py
import os
from pathlib import Path

from akerbp.mlops.core import logger


logging=logger.get_logger(name='mlops_core')


def get_top_folder(path):
    """
    Get the top folder of a path.

    Input:
        - path: (str or Path) 

    Output:
        - (str or Path, depending on input) top parent folder in path
    """
    if isinstance(path, str):
        return path.split(os.sep)[0]
    elif isinstance(path, Path):
        return path.parents[len(path.parents)-2]


def as_import_path(file_path):
    if file_path:
        if not isinstance(file_path, str):
            file_path=str(file_path)
        return file_path.replace(os.sep,'.').replace('.py','')
    else:
        logging.debug(f"Empty file path -> empty import path returned")


def confirm_prompt(question):
    reply = None
    while reply not in ("y", "n"):
        reply = input(f"{question} (y/n): ").lower()
    return (reply == "y")
